import torch
from .MultiHeadGraphAttentionModule import MultiHeadGraphAttentionModule


class GraphAttentionModule(torch.nn.Module):
    def __init__(
        self,
        input_dimension: int,
        hidden_dimension: int,
        output_dimension: int,
        num_heads: int,
        dropout: float,
        alpha: float,
        concat: bool,
    ):
        super().__init__()
        self.theta = MultiHeadGraphAttentionModule(
            input_dimension, hidden_dimension, num_heads, dropout, alpha, concat
        )
        self.sigma = MultiHeadGraphAttentionModule(
            hidden_dimension * num_heads, output_dimension, 1, dropout, alpha, concat
        )
        self.dropout = torch.nn.Dropout(dropout)

    def forward(self, features: torch.Tensor, adjacency_matrix: torch.Tensor):
        features = self.dropout(features)
        features = self.theta(features, adjacency_matrix)
        features = self.dropout(features)
        features = self.sigma(features, adjacency_matrix)
        return torch.nn.functional.log_softmax(features, dim=1)



import torch
from .modules import GraphAttentionModule
from ..core.concerns import MovesModulesToDevice


class GraphAttention(MovesModulesToDevice):
    def __init__(
        self,
        input_dimension: int,
        output_dimension: int,
        hidden_dimension: int = 8,
        num_heads: int = 8,
        alpha: float = 0.2,
        lr: float = 0.005,
        decay: float = 0.0005,
        dropout: float = 0.6,
        concat: bool = True,
    ):
        super().__init__()

        self.network = GraphAttentionModule(
            input_dimension=input_dimension,
            output_dimension=output_dimension,
            hidden_dimension=hidden_dimension,
            num_heads=num_heads,
            dropout=dropout,
            alpha=alpha,
            concat=concat,
        )
        self.optimizer = torch.nn.AdamW(
            self.network.parameters(), lr=lr, weight_decay=decay
        )
        self.criterion = torch.nn.CrossEntropyLoss()
        self.move_modules()



import torch


class GraphAttentionHeadModule(torch.nn.Module):
    def __init__(
        self, input_dimension: int, output_dimension: int, dropout: float, alpha: float
    ):
        super().__init__()
        # feature transform
        self.phi = torch.nn.Linear(input_dimension, output_dimension, bias=False)

        # attention for source
        self.mu = torch.nn.Linear(output_dimension, 1, bias=False)

        # attention for target
        self.xi = torch.nn.Linear(output_dimension, 1, bias=False)

        self.leakyrelu = torch.nn.LeakyReLU(alpha)
        self.dropout = torch.nn.Dropout(dropout)

    def forward(self, features: torch.Tensor, adjacency_matrix: torch.Tensor):
        # transform features
        h = self.phi(features)

        # compute pairwise scores
        e = self.mu(h) + self.xi(h).T
        e = self.leakyrelu(e)
        # masks non-edges
        e = torch.where(adjacency_matrix > 0, e, torch.full_like(e, -9e15))

        # normalize
        a = torch.nn.functional.softmax(e, dim=1)
        a = self.dropout(a)

        # aggregate neighbors
        return torch.nn.functional.elu(a @ h)



import torch
from ...device import move


class MovesModulesToDevice:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def move_modules(self):
        for module in dir(self):
            if module.startswith("__") or module.endswith("__"):
                continue

            attribute = getattr(self, module)
            if isinstance(attribute, torch.nn.Module):
                move(attribute)



import torch
from .GraphAttentionHeadModule import GraphAttentionHead


class MultiHeadGraphAttentionModule(torch.nn.Module):
    def __init__(
        self,
        input_dimension: int,
        output_dimension: int,
        num_heads: int,
        dropout: float,
        alpha: float,
        concat: bool,
    ):
        super().__init__()
        self.heads = torch.nn.ModuleList(
            [
                GraphAttentionHead(input_dimension, output_dimension, dropout, alpha)
                for _ in range(num_heads)
            ]
        )
        self.concat: bool = concat

    def forward(self, features: torch.Tensor, adjacency_matrix: torch.Tensor):
        features = [head(features, adjacency_matrix) for head in self.heads]

        # aggregate
        if self.concat:
            features = torch.cat(features, dim=1)
        else:
            features = torch.mean(torch.stack(features), dim=0)

        return features



import torch


class SoftActorModule(torch.nn.Module):
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        action_dimension: int,
        min_std: float,
        max_std: float,
    ):
        super().__init__()

        self.state_dimension: int = state_dimension
        self.hidden_dimension: int = hidden_dimension
        self.action_dimension: int = action_dimension
        self.min_std: float = float(min_std)
        self.max_std: float = float(max_std)

        self.phi = torch.nn.Sequential(
            torch.nn.Linear(self.state_dimension, self.hidden_dimension),
            torch.nn.ReLU(),
            torch.nn.Linear(self.hidden_dimension, self.hidden_dimension),
            torch.nn.ReLU(),
        )

        self.mu = torch.nn.Linear(self.hidden_dimension, self.action_dimension)
        torch.nn.init.uniform_(self.mu.weight, -3e-3, 3e-3)
        torch.nn.init.constant_(self.mu.bias, 0.0)

        self.xi = torch.nn.Linear(self.hidden_dimension, self.action_dimension)
        torch.nn.init.constant_(self.xi.weight, 0.0)
        torch.nn.init.constant_(self.xi.bias, -0.5)

    def forward(self, state: torch.Tensor):
        features = self.phi(state)
        mean = self.mu(features)
        std = self.xi(features).clamp(self.min_std, self.max_std)
        return mean, std
